@@ -329,7 +329,7 @@ static char *rcsid = "$Id: sk_g16.c,v 1.1 1994/06/30 16:25:15 root Exp $";
 #ifndef HAVE_PORTRESERVE
 
 #define check_region(ioaddr, size)              0
-#define register_iomem(ioaddr, size,name);             do ; while (0)
+#define request_region(ioaddr, size,name)       do ; while (0)
 
 #endif
 
@@ -487,7 +487,7 @@ static int   SK_probe(struct device *dev, short ioaddr);
 
 static int   SK_open(struct device *dev);
 static int   SK_send_packet(struct sk_buff *skb, struct device *dev);
-static void  SK_interrupt(int reg_ptr);
+static void  SK_interrupt(int irq, struct pt_regs * regs);
 static void  SK_rxintr(struct device *dev);
 static void  SK_txintr(struct device *dev);
 static int   SK_close(struct device *dev);
@@ -721,7 +721,7 @@ int SK_probe(struct device *dev, short ioaddr)
 
 	outb(SK_ROM_RAM_OFF, SK_POS2);     /* Boot_ROM + RAM off */  
 
-        /* We found a Boot_ROM and its gone. Set RAM address on
+        /* We found a Boot_ROM and it's gone. Set RAM address on
          * Boot_ROM address. 
          */ 
 
@@ -782,7 +782,7 @@ int SK_probe(struct device *dev, short ioaddr)
 	    dev->dev_addr[5]);
 
     /* Grab the I/O Port region */
-    register_iomem(ioaddr, ETHERCARD_TOTAL_SIZE,"sk_g16");
+    request_region(ioaddr, ETHERCARD_TOTAL_SIZE,"sk_g16");
 
     /* Initialize device structure */
 
@@ -1300,7 +1300,7 @@ static int SK_send_packet(struct sk_buff *skb, struct device *dev)
  * Description    : SK_G16 interrupt handler which checks for LANCE
  *                  Errors, handles transmit and receive interrupts
  *
- * Parameters     : I : int reg_ptr -
+ * Parameters     : I : int irq, struct pt_regs * regs -
  * Return Value   : None
  * Errors         : None
  * Globals        : None
@@ -1309,9 +1309,8 @@ static int SK_send_packet(struct sk_buff *skb, struct device *dev)
  *     YY/MM/DD  uid  Description
 -*/
 
-static void SK_interrupt(int reg_ptr)
+static void SK_interrupt(int irq, struct pt_regs * regs)
 {
-    int irq = - (((struct pt_regs *)reg_ptr)->orig_eax+2);
     int csr0;
     struct device *dev = (struct device *) irq2dev_map[irq];
     struct priv *p = (struct priv *) dev->priv;
